#ifndef __entity_def__
#define __entity_def__
#include <cassert>
#include <cstdlib>
#include "global.h"
#include "types.h"
#include "generic.h"
#include "random.h"
const int MaxE = 4096;

class Entity;

void init_entity (void);
void create_entity (Entity * par, int num,
 int d_x, int d_y, int v_x = 0, int v_y = 0);
void remove_entity (Entity * ent);

class Entity
{
private:
 int next_x, next_y;
 int dx, dy, vx, vy;
 Generic * type;
 Entity * next;
 bool primary, isrotor, issteam, issmoke;
 int effort, phase;
 int smoke_x [MaxSmoke], smoke_y [MaxSmoke], smoke_num;
public:
 bool is_primary (void) {return primary;}
 bool is_rotor (void) {return isrotor;}
 bool is_steam (void) {return issteam;}
 bool is_smoke (void) {return issmoke;}
 int getX (void) {return dx;}
 int getY (void) {return dy;}
 int getVX (void) {return vx;}
 int getVY (void) {return vy;}
 Entity * getNext (void) {return next;}
 Generic * getType (void) {return type;}
// void setType (Generic * n_type) {type = n_type;}

 void gen_steam (void)
 {
  int i, t;
  t = MinPhase + (effort << E_Scale) + abs (vx) + abs (vy);
  for (i = 0; i < type -> steam_num; i++)
   if (rndvalue (MaxPhase) <= t)
    create_entity (NULL, G_STEAM_1, dx + type -> steam_x[i],
     dy + type -> steam_y[i], vx, vy);
 }
 
 void add_phase (void)
 {
  if (issteam || issmoke)
  {
   phase += (MaxPhase >> ((SmokeTime + 1))) +
    rndvalue (MaxPhase >> SmokeTime);
   if (phase >= MaxPhase) {remove_entity (this); return;}
  }
  else
  {
   phase += MinPhase + (effort << E_Scale) + abs (vx) + abs (vy);
   phase &= MaxPhase - 1;
  }
  if (isrotor) type = generic[G_ROTOR_1 + ((phase * 10) >> P_Scale)]; else
  if (issteam) type = generic[G_STEAM_1 + ((phase * 5) >> P_Scale)]; else
  if (issmoke) type = generic[G_SMOKE_1 + ((phase * 5) >> P_Scale)];
//  printf ("%s %d\n", type -> name, phase);
 }
 
 void control_left (void)
 {
  effort = 1;
  const int & vel_left = type -> vel_left;
  const int & acc_left = type -> acc_left;
  const int & brk_left = type -> brk_left;
  if (vel_left)
  {
   if (vx > 0) {vx -= brk_left; if (vx < 0)         vx = 0;}
          else {vx -= acc_left; if (vx < -vel_left) vx = -vel_left;}
  }
  else dx -= acc_left;
 }

 void control_right (void)
 {
  const int & vel_right = type -> vel_right;
  const int & acc_right = type -> acc_right;
  const int & brk_right = type -> brk_right;
  if (vel_right)
  {
   if (vx < 0) {vx += brk_right; if (vx > 0)         vx = 0;}
          else {vx += acc_right; if (vx > vel_right) vx = vel_right;}
  }
  else dx += acc_right;
 }

 void control_up (void)
 {
  const int & vel_up = type -> vel_up;
  const int & acc_up = type -> acc_up;
  const int & brk_up = type -> brk_up;
  if (vel_up)
  {
   if (vy > 0) {vy -= brk_up; if (vy < 0)       vy = 0;}
          else {vy -= acc_up; if (vy < -vel_up) vy = -vel_up;}
  }
  else dy -= acc_up;
 }

 void control_down (void)
 {
  const int & vel_down = type -> vel_down;
  const int & acc_down = type -> acc_down;
  const int & brk_down = type -> brk_down;
  if (vel_down)
  {
   if (vy < 0) {vy += brk_down; if (vy > 0)        vy = 0;}
          else {vy += acc_down; if (vy > vel_down) vy = vel_down;}
  }
  else dy += acc_down;
 }
 
 void draw (void)
 {
  type -> draw (dx, dy);
  if (next != NULL)
   next -> draw ();
 }

 void move_next (void)
 {
  if (next != NULL)
  {
   next -> effort = effort;
   next -> dx = dx + next_x;
   next -> dy = dy + next_y;
   next -> vx = vx;
   next -> vy = vy;
   next -> move_next ();
  }
 }

 void move (void)
 {
  dx += vx; dy += vy;
  move_next ();
  effort = 0;
 }

 Entity (Entity * par, int num,
  int d_x, int d_y, int v_x = 0, int v_y = 0)
 {
  primary = (par == NULL);
  if (!primary)
  {
   par -> next = this;
   par -> next_x = d_x << H_Scale;
   par -> next_y = d_y << V_Scale;
  }
  type = generic[num];
  isrotor = (num >= G_ROTOR_1 && num <= G_ROTOR_5);
  issteam = (num >= G_STEAM_1 && num <= G_STEAM_5);
  issmoke = (num >= G_SMOKE_1 && num <= G_SMOKE_5);
  dx = d_x; dy = d_y; vx = v_x; vy = v_y;
  next = NULL;
  effort = 0;
  phase = 0;
 }

 ~Entity () {}
};

extern Entity * entity [MaxE];
extern int e_prev [MaxE], e_next [MaxE];
extern int e_full, e_free;
#endif // __entity_def__
