#include <cassert>
#include <cmath>
#include <cstdio>
#include <cstring>
#include "global.h"
#include "random.h"
#include "types.h"
#include "generic.h"
#include "entity.h"

int colour [COLOUR_NUM];

char tbuf [SizeTY] [SizeTX];
byte zbuf [SizeTY] [SizeTX];
unsigned int tcol [SizeTY] [SizeTX];

BITMAP * gbuf [2];
int gb0, gb1;
bool kb_poll;
bool End1, win, lose;

int camera_x, camera_y;
int camera_dx, camera_dy;
int camera_ptr, train_ptr;
int tree_seed, cloud_seed, interceptor_seed, bomber_seed, fighter_seed;
int time_left;

int rail [MaxRail];
int tree [MaxRail];

inline int t2x (int v) {return v << 3;}
inline int x2t (int v) {return v >> 3;}
inline int t2y (int v) {return v << 3;}
inline int y2t (int v) {return v >> 3;}

void draw (void);

inline void poll_kb (void)
{
 if (kb_poll) poll_keyboard ();
}

void try_init (bool res, const char * str)
{
 if (res) printf ("%s init OK.\n", str);
 else
 {
  printf ("%s init failed!\n", str);
  exit (EXIT_FAILURE);
 }
}

void init_kb (void)
{
 try_init (install_keyboard () != -1, "Keyboard");
 kb_poll = keyboard_needs_poll ();
}

void init (void)
{
 FILE * f;
 int k;

 try_init (allegro_init () == 0, "Allegro");
 init_kb ();
 set_uformat (U_ASCII);
 set_color_depth (32);
 try_init (set_gfx_mode (GFX_AUTODETECT_WINDOWED, MaxX, MaxY, 0, 0) == 0,
  "Graphics");

 gb0 = 0; gb1 = 1;
 gbuf[0] = create_video_bitmap (MaxX, MaxY); assert (gbuf[0] != NULL);
 gbuf[1] = create_video_bitmap (MaxX, MaxY); assert (gbuf[1] != NULL);

 colour[COLOUR_BACK] = 0x00000000;
 colour[COLOUR_FORE] = 0x004080B0;
 colour[COLOUR_HELP] = 0x0040B080;
 colour[COLOUR_DANGER] = 0x00D02020;
 colour[COLOUR_OK] = 0x0020D020;
 
 tree_seed = 100;
 cloud_seed = 100;
 interceptor_seed = 12;
 bomber_seed = 9;
 fighter_seed = 3;
 
 if ((f = fopen (frnd_name, "rt")) == NULL)
 {
  k = 0;
 }
 else
 {
  fscanf (f, "%d", &k);
  fclose (f);
 }
 f = fopen (frnd_name, "wt");
 fprintf (f, "%d\n", k + 1);
 fclose (f);
 if (!initrand (k)) exit (EXIT_FAILURE);
 printf ("Started with random init %d.\n", k);
 fflush (stdout);

 for (k = 0; k < GENERIC_NUM; k++)
  generic[k] = new Generic (k);

 camera_dx = -15;
 camera_dy = -10;
 init_entity ();
}

void happy_end (void)
{
 int k;
 while (e_full != -1)
  remove_entity (entity[e_full]);
 for (k = 0; k < GENERIC_NUM; k++)
  delete generic[k];
 destroy_bitmap (gbuf[0]);
 destroy_bitmap (gbuf[1]);
 allegro_exit ();
 printf ("Successful termination.\n");
 fflush (stdout);
 exit (EXIT_SUCCESS);
}

void display_help (void)
{
 int kb, kc;
 gb0 ^= 1; gb1 ^= 1;
 BITMAP * gcur = gbuf[gb0];
 acquire_bitmap (gcur);
 clear_to_color (gcur, colour[COLOUR_BACK]);
 textout_centre_ex (gcur, font, "Zeitnot",
  t2x (TextX >> 1), t2y ((TextY >> 1) - 15), colour[COLOUR_HELP], -1);
 textout_centre_ex (gcur, font, "You control the helicopter.",
  t2x (TextX >> 1), t2y ((TextY >> 1) - 9), colour[COLOUR_HELP], -1);
 textout_centre_ex (gcur, font, "The goal is to get the train "
  "to the next station in time.",
  t2x (TextX >> 1), t2y ((TextY >> 1) - 7), colour[COLOUR_HELP], -1);
 textout_centre_ex (gcur, font, "Advance to make the train go faster.",
  t2x (TextX >> 1), t2y ((TextY >> 1) - 5), colour[COLOUR_HELP], -1);
 textout_centre_ex (gcur, font, "Fall back to make it go slower.",
  t2x (TextX >> 1), t2y ((TextY >> 1) - 3), colour[COLOUR_HELP], -1);
 textout_centre_ex (gcur, font, "It's currently a peaceful teaser only, "
  "battle features are not yet implemented.",
  t2x (TextX >> 1), t2y ((TextY >> 1) + 0), colour[COLOUR_DANGER], -1);
 textout_centre_ex (gcur, font, "Nothing can yet prevent you "
  "from acheiving the goal.",
  t2x (TextX >> 1), t2y ((TextY >> 1) + 2), colour[COLOUR_DANGER], -1);
 textout_centre_ex (gcur, font, "Use the arrow keys to control "
  "the helicopter.",
  t2x (TextX >> 1), t2y ((TextY >> 1) + 5), colour[COLOUR_HELP], -1);
 textout_centre_ex (gcur, font, "Press ESC or Q to quit.",
  t2x (TextX >> 1), t2y ((TextY >> 1) + 7), colour[COLOUR_HELP], -1);
 release_bitmap (gcur);
 show_video_bitmap (gcur);
 while (!keypressed ()) ;
 while (keypressed ()) kb = ureadkey (&kc);
}

void display_win (void)
{
 int kb, kc;
 gb0 ^= 1; gb1 ^= 1;
 BITMAP * gcur = gbuf[gb0];
 acquire_bitmap (gcur);
 textout_centre_ex (gcur, font, "              ",
  t2x (TextX >> 1), t2y ((TextY >> 1) - 2), colour[COLOUR_OK], -1);
 textout_centre_ex (gcur, font, "              ",
  t2x (TextX >> 1), t2y ((TextY >> 1) - 1), colour[COLOUR_OK], -1);
 textout_centre_ex (gcur, font, "   YOU WIN!   ",
  t2x (TextX >> 1), t2y ((TextY >> 1) + 0), colour[COLOUR_OK], -1);
 textout_centre_ex (gcur, font, "              ",
  t2x (TextX >> 1), t2y ((TextY >> 1) + 1), colour[COLOUR_OK], -1);
 textout_centre_ex (gcur, font, "              ",
  t2x (TextX >> 1), t2y ((TextY >> 1) + 2), colour[COLOUR_OK], -1);
 release_bitmap (gcur);
 show_video_bitmap (gcur);
 while (!keypressed ()) ;
 while (keypressed ()) kb = ureadkey (&kc);
}

void display_lose (void)
{
 int kb, kc;
 gb0 ^= 1; gb1 ^= 1;
 BITMAP * gcur = gbuf[gb0];
 acquire_bitmap (gcur);
 textout_centre_ex (gcur, font, "                 ",
  t2x (TextX >> 1), t2y ((TextY >> 1) - 2), colour[COLOUR_DANGER], -1);
 textout_centre_ex (gcur, font, "                 ",
  t2x (TextX >> 1), t2y ((TextY >> 1) - 1), colour[COLOUR_DANGER], -1);
 textout_centre_ex (gcur, font, "   YOU LOSE...   ",
  t2x (TextX >> 1), t2y ((TextY >> 1) + 0), colour[COLOUR_DANGER], -1);
 textout_centre_ex (gcur, font, "                 ",
  t2x (TextX >> 1), t2y ((TextY >> 1) + 1), colour[COLOUR_DANGER], -1);
 textout_centre_ex (gcur, font, "                 ",
  t2x (TextX >> 1), t2y ((TextY >> 1) + 2), colour[COLOUR_DANGER], -1);
 release_bitmap (gcur);
 show_video_bitmap (gcur);
 while (!keypressed ()) ;
 while (keypressed ()) kb = ureadkey (&kc);
}

void control_player (int k)
{
 int kb, kc;
 Entity * & cur = entity[k];
 poll_kb ();
 if (key[KEY_LEFT ] || key[KEY_4_PAD]) cur -> control_left ();
 if (key[KEY_RIGHT] || key[KEY_6_PAD]) cur -> control_right ();
 if ((cur -> getY () >> V_Scale) > 0)
  if (key[KEY_UP   ] || key[KEY_8_PAD]) cur -> control_up ();
 if ((cur -> getY () >> V_Scale) < 32)
  if (key[KEY_DOWN ] || key[KEY_2_PAD]) cur -> control_down ();
 while (keypressed ())
 {
  kb = ureadkey (&kc);
  switch (kc)
  {
   case KEY_F1:
    display_help ();
    break;
   case KEY_ESC:
   case KEY_Q:
    End1 = true;
    break;
  }
 }
 while (keypressed ()) kb = ureadkey (&kc);
}

void control_train (int k)
{
 int i, j, q, t;
 Entity * & cur = entity[k];
 t = cur -> getType () -> get_max_vel ();
 q = t >> 2;
 if (cur -> getVX () > 0)
  for (i = 0, j = i + ((cur -> getX ()) >> H_Scale); i < 75; i++)
   if (rail[i] > G_R_LEFT)
   {cur -> control_left (); return;}
 if (cur -> getVX () < q)
 {cur -> control_right (); return;}
 if (cur -> getX () + (10 << H_Scale) < entity[camera_ptr] -> getX ())
 {cur -> control_right (); return;}
 else if (cur -> getX () - (12 << H_Scale) > entity[camera_ptr] -> getX ())
 {cur -> control_left (); return;}
}

void control_steam (int k)
{
 int t;
 Entity * & cur = entity[k];
 t = cur -> getVX ();
 switch (rndvalue (10))
 {
  case 0:
   cur -> control_right ();
   break;
  case 1:
   cur -> control_left ();
   break;
  default:
   if (t < 0) cur -> control_right ();
   else if (t > 0) cur -> control_left ();
   break;
 }
 cur -> control_up ();
}

void control_smoke (int k)
{
 int t;
 Entity * & cur = entity[k];
 t = cur -> getVX ();
 switch (rndvalue (10))
 {
  case 0:
   cur -> control_right ();
   break;
  case 1:
   cur -> control_left ();
   break;
  default:
   if (t < 0) cur -> control_right ();
   else if (t > 0) cur -> control_left ();
   break;
 }
 cur -> control_up ();
}

void control (void)
{
 int i;
 for (i = e_full; i != -1; i = e_next[i])
  if (entity[i] -> is_primary ())
  {
   if (i == camera_ptr)          control_player (i); else
   if (i == train_ptr)           control_train (i); else
   if (entity[i] -> is_steam ()) control_steam (i); else
   if (entity[i] -> is_smoke ()) control_smoke (i);
  }
}

void move (void)
{
 int i;
 for (i = e_full; i != -1; i = e_next[i])
  if (entity[i] -> is_primary ())
   entity[i] -> move ();
}

void effects_player (int k)
{
 Entity * & cur = entity[k];
 cur -> getNext () -> add_phase ();
}

void effects_steam (int k)
{
 Entity * & cur = entity[k];
 cur -> add_phase ();
}

void effects_smoke (int k)
{
 Entity * & cur = entity[k];
 cur -> add_phase ();
}

void effects (void)
{
 int i, t;
 for (i = e_full; i != -1; i = t)
 {
  t = e_next[i];
  entity[i] -> gen_steam ();
  if (entity[i] -> is_primary ())
  {
   if (i == camera_ptr)          effects_player (i); else
   if (entity[i] -> is_steam ()) effects_steam (i); else
   if (entity[i] -> is_smoke ()) effects_smoke (i);
  }
 }
}

void draw (void)
{
 int i, j;
 char buf[2];
 buf[1] = 0;
 gb0 ^= 1; gb1 ^= 1;
 for (j = 0; j < SizeTY; j++)
  for (i = 0; i < SizeTX; i++)
   tbuf[j][i] = ' ';
 for (j = 0; j < SizeTY; j++)
  for (i = 0; i < SizeTX; i++)
   zbuf[j][i] = 0;
 for (j = 0; j < SizeTY; j++)
  for (i = 0; i < SizeTX; i++)
   tcol[j][i] = colour[COLOUR_FORE];

 for (i = MinTX - 6, j = (i << H_Scale) + camera_x; i < MaxTX + 7;
  i++, j += 1 << H_Scale)
  if (tree[j >> H_Scale])
   generic[tree[j >> H_Scale]] -> draw (j, 30 << V_Scale);
 for (i = MinTX, j = (i << H_Scale) + camera_x; i < MaxTX;
  i++, j += 1 << H_Scale)
  generic[rail[j >> H_Scale]] -> draw (j, 35 << V_Scale);
 for (i = e_full; i != -1; i = e_next[i])
  if (entity[i] -> is_primary ())
   entity[i] -> draw ();

 BITMAP * gcur = gbuf[gb0];
 acquire_bitmap (gcur);
 clear_to_color (gcur, colour[COLOUR_BACK]);
 textprintf_ex (gcur, font, t2x ((TextX >> 1) - 7), t2y (3),
  colour[COLOUR_FORE], -1, "Time left: %4d", time_left);
 textprintf_ex (gcur, font, t2x ((TextX >> 1) - 9), t2y (5),
  colour[COLOUR_FORE], -1, "Distance left: %4d",
   EndRail - (entity[train_ptr] -> getX () >> H_Scale));
 for (j = MinTY; j < MaxTY; j++)
  for (i = MinTX; i < MaxTX; i++)
  {
   buf[0] = tbuf[j - MinTY][i - MinTX];
   textout_ex (gcur, font, buf,
    t2x (i), t2y (j), tcol[j - MinTY][i - MinTX], -1);
  }
 release_bitmap (gcur);
 show_video_bitmap (gcur);
}

void init_rail (void)
{
 int k;
 for (k = 0; k < StartRail - PartRail; k++)
  rail[k] = G_R_HANGAR;
 for (k = StartRail - PartRail; k < StartRail; k++)
  rail[k] = (k % 7) ? G_R_PLATFORM : G_R_LIGHT;
 for (k = StartRail; k < EndRail; k++)
  rail[k] = G_R_SOLID;
 for (k = EndRail; k < EndRail + PartRail; k++)
  rail[k] = (k % 7) ? G_R_PLATFORM : G_R_LIGHT;
 for (k = EndRail + PartRail; k < MaxRail; k++)
  rail[k] = G_R_HANGAR;
 assert (G_TREE_1);
 for (k = 0; k < MaxRail; k++)
  tree[k] = 0;
 for (k = 0; k < tree_seed; k++)
  tree[rndvalue (EndRail - StartRail + 1) + StartRail] =
   G_TREE_1 + rndvalue (2);
 time_left = MaxTime;

 create_entity (NULL, G_LOCOMOTIVE, StartRail << H_Scale, 33 << V_Scale, 0, 0);
 train_ptr = e_full;
 create_entity (entity[e_full],        G_TENDER, -10, 3);
 create_entity (entity[e_full],      G_SIGN_CAR, -12, 1);
 create_entity (entity[e_full],      G_FUEL_CAR, -12, 0);
 create_entity (entity[e_full],      G_AMMO_CAR, -12, 0);
 create_entity (entity[e_full],      G_LAND_CAR, -12, 0);
 create_entity (entity[e_full],    G_REPAIR_CAR, -12, 0);
 create_entity (entity[e_full], G_PASSENGER_CAR, -12, 0);
 create_entity (NULL, G_HELICOPTER, StartRail << H_Scale, 20 << V_Scale, 0, 0);
 camera_ptr = e_full;
 create_entity (entity[e_full], G_ROTOR_1, 0, -2);

 for (k = 0; k < cloud_seed; k++)
 {
  create_entity (NULL, G_CLOUD_1 + rndvalue (3),
   (rndvalue (EndRail + StartRail + (StartRail >> 1) + 1) +
   (StartRail >> 1)) << H_Scale, (rndvalue (11) << V_Scale) - 5,
   (rndvalue (1 << (H_Scale - 2)) + (1 << (H_Scale - 2))) *
   (rndvalue (2) ? 1 : -1), 0);
 }

 for (k = 0; k < interceptor_seed; k++)
 {
  create_entity (NULL, G_INTERCEPTOR,
   (rndvalue ((EndRail << 1) + StartRail) + (StartRail << 1)) << H_Scale,
   (rndvalue (10) << V_Scale) + 3,
   -(generic[G_INTERCEPTOR] -> get_max_vel () >> 1), 0);
 }

 for (k = 0; k < bomber_seed; k++)
 {
  create_entity (NULL, G_BOMBER,
   (rndvalue ((EndRail << 1) + StartRail) + (StartRail << 1)) << H_Scale,
   (rndvalue (5) << V_Scale) + 8,
   -(generic[G_BOMBER] -> get_max_vel () >> 1), 0);
 }

 for (k = 0; k < fighter_seed; k++)
 {
  create_entity (NULL, G_FIGHTER,
   (rndvalue ((EndRail << 1) + StartRail) + (StartRail << 1)) << H_Scale,
   (rndvalue (15) << V_Scale) + 4,
   -(generic[G_FIGHTER] -> get_max_vel () >> 1), 0);
 }
}

void clear_rail (void)
{
 int i, t;
 for (i = e_full; i != -1; i = t)
 {
  t = e_next[i];
  if (entity[i] -> is_primary ())
   remove_entity (entity[i]);
 }
}

void game (void)
{
 init_rail ();
 End1 = win = lose = false;
 while (!End1)
 {
  move ();
  control ();
  effects ();
  camera_x = entity[camera_ptr] -> getX () + 
   entity[camera_ptr] -> getVX () * 12 + (camera_dx << H_Scale);
  camera_y = camera_dy << V_Scale;
  draw ();
  time_left--;
  if ((entity[train_ptr] -> getX () >> H_Scale) > EndRail) win = End1 = true;
  if (time_left <= 0) lose = End1 = true;
  rest (delay_time);
 }

 if (win) display_win ();
 else if (lose) display_lose ();

 clear_rail ();
}

void display_menu (void)
{
 gb0 ^= 1; gb1 ^= 1;
 BITMAP * gcur = gbuf[gb0];
 acquire_bitmap (gcur);
 clear_to_color (gcur, colour[COLOUR_BACK]);
 textout_centre_ex (gcur, font, "Zeitnot",
  t2x (TextX >> 1), t2y ((TextY >> 1) - 9), colour[COLOUR_FORE], -1);
 textout_centre_ex (gcur, font, "version 0.2",
  t2x (TextX >> 1), t2y ((TextY >> 1) - 7), colour[COLOUR_FORE], -1);
 textout_centre_ex (gcur, font, "by Gassa",
  t2x (TextX >> 1), t2y ((TextY >> 1) - 5), colour[COLOUR_FORE], -1);
 textout_centre_ex (gcur, font, "Press Enter to start",
  t2x (TextX >> 1), t2y ((TextY >> 1) + 2), colour[COLOUR_FORE], -1);
 textout_centre_ex (gcur, font, "Press Esc or Q to quit",
  t2x (TextX >> 1), t2y ((TextY >> 1) + 4), colour[COLOUR_FORE], -1);
 textout_centre_ex (gcur, font, "Press F1 to show the help screen",
  t2x (TextX >> 1), t2y ((TextY >> 1) + 6), colour[COLOUR_FORE], -1);
 release_bitmap (gcur);
 show_video_bitmap (gcur);
}

void menu (void)
{
 int kb, kc;
 bool EndMenu;
 EndMenu = false;
 while (!EndMenu)
 {
  rest (delay_time);
  display_menu ();
  poll_kb ();
  while (keypressed ())
  {
   kb = ureadkey (&kc);
   switch (kc)
   {
    case KEY_F1:
     display_help ();
     break;
    case KEY_ENTER:
     game ();
     break;
    case KEY_ESC:
    case KEY_Q:
     EndMenu = true;
     break;
   }
  }
 }
}

int main (void)
{
 init ();
 menu ();
 happy_end ();
 return 0;
}
END_OF_MAIN ()
